function varargout = phipade(z, k)
% PHIPADE - Evaluate phi functions using (6,6)-Pad approximations.
%
% SYNOPSIS:
%    phi_k                     = phipade(z, k);
%   [phi_1, phi_2, ..., phi_k] = phipade(z, k);
%
% DESCRIPTION:
%   This function evaluates phi functions needed in exponential
%   integrators using diagonal Pad approximants (currently (6,6)-Pad).
%   We define the phi functions according to the integral representation
%
%      \phi_k(z) = \frac{1}{(k - 1)!} \int_0^1 e^{z (1-x)} x^{k-1} dx
%
%   for k=1, 2, ...
%
% PARAMETERS:
%   z - Evaluation point.  Assumed to be one of
%         - 1D vector, treated as the main diagonal of a diagonal matrix
%         - sparse diagonal matrix
%         - full or sparse matrix
%   k - Which phi function(s) to evaluate.
%       Index (integer) of the (highest) phi function needed.
%
% RETURNS:
%    phi_k                     =      \phi_k(z)
%   [phi_1, phi_2, ..., phi_k] = DEAL(\phi_1(z), \phi_2(z), ..., \phi_k(z))
%
% NOTES:
%   When computing more than one phi function, it is the caller's
%   responsibility to provide enough output arguments to hold all of the
%   \phi_k function values.
%
%   For efficiency reasons, PHIPADE caches recently computed function
%   values.  The caching behaviour is contingent on the WANTCACHE
%   function and may be toggled on or off as needed.
%
% SEE ALSO:
%   WANTCACHE.

% This file is part of the 'Expint'-package,
% see http://www.math.ntnu.no/num/expint/
%
% $Revision: 1.13 $  $Date: 2005/05/15 15:49:44 $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Local user configuration.
CACHE_BND = 4;          % upper bound on number of cache entries
atol = 1.0e-12;         % FLTEQ absolute error tolerance
rtol = 5.0e-13;         % FLTEQ relative error tolerance

% Uncomment to remove WANTCACHE function dependency (ie. make PHIPADE
% run in a self contained environment).
% wantcache = true;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% No user changeable parameters below this line.

% Note: Short circuiting operators are critical in this statement.  In
%       particular we need the NUMEL check before the k==NARGOUT check
%       because the operators only accept scalar (logical) operands.
%
%       We furthermore note the inconsistency of NARGOUT.  NARGOUT is a
%       peculiar function and statements of the form
%
%               nargout (op) <some number>
%
%       generates errors due to too many input args to NARGOUT.
%       However, such statements *ARE* allowed within IF statements...

arg_ok = (0 == nargout) || (numel(k) == nargout) || (k == nargout);
if ~arg_ok,
   error('EXPINT:phipade:nargout', ...
         'Inconsistent number of output arguments');
end

% support the
%        [phi{1:k}] = phipade(z, k);
% syntax
if (nargout > 1) && (numel(k) == 1), k = 1:k; end

% treat vectors as (sparse) diagonal matrices
if sum(size(z) > 1) == 1,
   n = length(z);
   z = spdiags(z(:), 0, n, n);
end

if wantcache,
   % main cache data structure
   % see *_cache() functions for operational detail...
   persistent phi_cache;

   idx = find_cache(phi_cache, z, max(k), atol, rtol);
   if idx < 1,
      [idx, phi_cache] = put_cache(phi_cache, z, max(k), idx);
   end

   nk = numel(k);
   [phi_cache, varargout{1:nk}] = get_cache(phi_cache, idx, k, CACHE_BND);
else
   pv = eval_pade(z, max(k));
   [varargout{1:numel(k)}] = deal(pv{1}{k});
end


function idx = find_cache(phi, z, k, atol, rtol)
idx = 0;

if ~isempty(phi)
   nrm_z = norm(z, inf);
   siz   = size(z);
   siz_c = numel(phi.LRU.list);

   % likely non-optimal, but use linear search for now...
   j = 0;
   while (idx == 0) && (j < siz_c),
      j = j + 1;
      p = phi.LRU.list(phi.LRU.idx(j));

      if match_cache(p, z, nrm_z, siz, atol, rtol),
         if p.maxk >= k,
            idx = j;            % all requirements satisfied
         else
            idx = -j;           % correct z, too few phi's
         end
      end
   end

   if idx == 0,
      j = match_scaled(phi.scaled, z, k, atol, rtol);

      % relies on sign(j)==0 for j==0
      idx = sign(j) * (siz_c + abs(j));
   end
end


function [phi, varargout] = get_cache(phi, idx, k, bound)
% - idx > 0
% - numel(phi.LRU.list) > 0
% - idx < numel(phi.LRU.list) || ~isempty(phi.scaled(end).z)

[nl, ns] = deal(numel(phi.LRU.list), numel(phi.scaled));
varargout = cell([1, numel(k)]);
if idx <= nl + ns;
   % request directly satisfied from phi.LRU or phi.scaled

   if idx <= nl,
      p = phi.LRU.list(phi.LRU.idx(idx));
      phi.LRU.idx = phi.LRU.idx([idx, 1:idx-1, idx+1:nl]);
   else
      idx = idx - nl;
      p = phi.scaled(idx);
      phi.LRU.list = [phi.scaled(idx), phi.LRU.list];
      phi.LRU.idx  = [1, 1 + phi.LRU.idx];
   end

   siz  = p.siz;
   maxk = p.maxk;

   phi_vals = mat2cell(p.phi, siz(1), repmat(siz(2), [1, maxk]));
else
   % z = 2^p * phi.scaled(end).z, but outside of phi.scaled
   % must do a bit of squaring

   s  = idx - (nl + ns) + 1;    % number of squarings needed
   p  = phi.scaled(1);

   siz  = p.siz;
   maxk = p.maxk;
   Id   = speye(siz);
   pv   = mat2cell(p.phi, siz(1), repmat(siz(2), [1, maxk]));

   [phi_vals, d{1:s}] = square_pade(p.z, Id, s, pv);
   z = 2^s .* p.z;

   phi.LRU.list = [struct('siz', siz,            ...
                          'nrm_z', norm(z, inf), ...
                          'maxk', max(k),        ...
                          'z', z,                ...
                          'phi', cat(2, phi_vals{:})), ...
                   phi.LRU.list];
   phi.LRU.idx  = [1, 1 + phi.LRU.idx];
end

% maintain upper bound on number of cached entries
if numel(phi.LRU.list) > bound,
   phi.LRU.list = phi.LRU.list(phi.LRU.idx(1:bound));
   phi.LRU.idx  = 1:bound;      % data copying normalises LRU index
end

% return values
[varargout{1:numel(k)}] = deal(phi_vals{k});


function [idx, phi] = put_cache(phi, z, k, idx)
% - request not satisfied from cache, do complete eval, set phi.scaled
% - further optimisations are possible in this case but not without
%   adversely affecting code readability...

siz  = size(z);
maxk = max(k);

phi_list = eval_pade(z, k);

s = struct('siz', siz, 'nrm_z', norm(z, inf), ...
           'maxk', maxk, 'z', z, 'phi', cat(2, phi_list{1}{:}));

phi.scaled = struct('siz', 0, 'nrm_z', 0, 'maxk', 0, 'z', [], 'phi', []);
for j = 2:numel(phi_list),
   z = z / 2;
   phi.scaled(j-1) = struct('siz',   siz,          ...
                            'nrm_z', norm(z, inf), ...
                            'maxk',  maxk,         ...
                            'z', z, 'phi', cat(2, phi_list{j}{:}));
end

if isempty(phi) || ~isfield(phi, 'LRU'),
   phi.LRU.list = [];
   phi.LRU.idx  = [];
end

idx = abs(idx);
n = numel(phi.LRU.list);
if (0 < idx) && (idx <= n),
   % - new phi-list computed for additional \phi_\ell functions
   % - replace existing phi.LRU z-entry

   phi.LRU.list(phi.LRU.idx(idx)) = s;
   phi.LRU.idx = phi.LRU.idx([idx, 1:idx-1, idx+1:n]);
else
   % - either not found or insufficient \phi_\ell functions
   % - prepend the newly computed phi functions to existing list

   phi.LRU.list = [s, phi.LRU.list];
   phi.LRU.idx  = [1, 1 + phi.LRU.idx];
end

idx = 1;        % entry found at phi.LRU.idx(1)


function phi_list = eval_pade(z, k)
[s, z] = scaled_arg(z);
Id = speye(size(z));
[phi_vals{1:k}] = scaled_pade(z, Id, 6, k);     % (6,6)-Pad approx

if s > 0,
   [phi_list{1:s+1}] = square_pade(z, Id, s, phi_vals);
else
   phi_list = {phi_vals};
end


function idx = match_scaled(phi, z, k, atol, rtol)
idx = 0;

if ~isempty(phi),
   [s, z] = scaled_arg(z);
   sgn = 1;

   if match_cache(phi(end), z, norm(z, inf), numel(z), atol, rtol),
      if phi(end).maxk < k, sgn = -1; end

      if s <= numel(phi),
         idx = numel(phi) - s;
      else
         idx = s;
      end

      idx = sgn * idx;
   end
end


function b = match_cache(p, z, nrm_z, siz, atol, rtol)
b = all(p.siz == siz)                 && ...
    flteq(p.nrm_z, nrm_z, atol, rtol) && ...
    flteq(p.z, z, atol, rtol);


function [s, z] = scaled_arg(z)
% Scaling to obtain well-behaved Pad approximations.
s = max(0, nextpow2(norm(z, inf)) + 1);
z = z ./ 2^s;


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Internal eval functions implementing Pad algorithm.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function varargout = scaled_pade(z, Id, d, m)
Z = z;
[num, den] = pade_cof(d, m);

% prealloc
N = cell([m, 1]);
D = cell([m, 1]);
varargout = cell([m, 1]);

% setup, assumes d >= 1, Id coeff generally ~= 1 (at least in numerator)
for ell = 1:m,
   N{ell} = num{ell}(1)*Id + num{ell}(2)*Z;
   D{ell} = den{ell}(1)*Id + den{ell}(2)*Z;
end

for j = 3:d+1,
   Z = Z * z;

   for ell = 1:m,
      N{ell} = N{ell} + num{ell}(j)*Z;
      D{ell} = D{ell} + den{ell}(j)*Z;
   end
end

% (d,d)-Pad approx of scaled argument
for ell = 1:m,
   varargout{ell} = D{ell} \ N{ell};
end


function varargout = square_pade(z, Id, s, phi)
% Undo scaling (squaring).
%
% Formulae derived by W. Wright.

[varargout{1:s+1}] = deal(phi);         % prealloc
Exp = z*phi{1} + Id;                    % exponential

for m = 1:s,
   varargout{end-m}{1} = (Exp + Id) * phi{1} / 2;

   i = [0, 1];
   p = ~i;
   for k = 2:numel(phi),
      i = i + p;
      v = phi{i(1)} * phi{i(2)};

      c = 2;
      a = mod(k, 2);
      ell = floor(k / 2);
      for j = k : -1 : ell + 1 + a,
         v = v + c.*phi{j};
         c = c / (k + 1 - j);
      end

      % odd-numbered phi's need special coeff in \phi_{\ell+1} term
      if a > 0, v = v + phi{ell+1}./prod(1:ell); end

      varargout{end-m}{k} = v / 2^k;
      p = ~p;
   end

   [phi{:}] = deal(varargout{end-m}{:});
   Exp = Exp * Exp;
end


function [N, D] = pade_cof(d, k)
% Re-normalised (d,d)-Pad coefficients for the \phi_\ell functions.
%
%%%% % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % % %
%
% This code is based on the explicit (d,d)-Pad formulae derived by W.
% Wright, but uses recurrence relations to reduce the computational
% complexity.  In particular, the coefficients n_i of the numerator
% polynomial N_d^\ell are defined by
%
%    n_i = \sum_{j=0}^i a_{ij} = sum(tril(A), 2)_i
%
% for i=0:d, in which
%
%    a_{ij} = (2d + ell - j)! (-1)^j / (j! (d-j)! (ell + i - j)!)
%           = -(d+1 - j) (ell+1+i - j) / (j (2d+ell+1 - j)) a_{i,j-1}
%
% for j=1:i.  Similar recurrence relations may be derived for the other
% coefficients, and for the denominator polynomial D_d^\ell.
%
% We note that roundoff errors unfortunately affects the accuracy of the
% coefficients.  However, as the errors are generally in the order of
% 1-5 ULP, we do not implement more accurate evaluation routines at this
% time.

n1 = prod(d + 1 : 2*d + 1);     % (2d + 1)! / d!
d1 = n1;

i = 1:d;
[J, I] = meshgrid(i);   % MESHGRID gives wrong order for this purpose

N = cell([k, 1]);
D = cell([k, 1]);
A = zeros(d + 1);

ell = 1;
while ell <= k,
   A(:, 1) = n1 .* cumprod([1, 1 ./ (ell + i)]) .';
   A(2:end, 2:end) = - (d + 1 - J) .* (ell + 1 + I - J) ./ ...
                       ((2*d + ell + 1 - J) .* J);

   N{ell} = sum(tril(cumprod(A, 2)), 2);
   D{ell} = d1 .* cumprod([1, -(d + 1 - i) ./ (i .* (2*d + ell + 1 - i))]) .';

   ell = ell + 1;

   n1 = n1 * (2*d + ell) / ell;
   d1 = d1 * (2*d + ell);
end


function B = flteq(x, y, varargin)
% FLTEQ - Determine floating point equality for arrays of DOUBLE or
%         COMPLEX.
%
% SYNOPSIS:
%   B = flteq(x, y);
%   B = flteq(x, y, atol);
%   B = flteq(x, y, atol, rtol);
%
% DESCRIPTION:
%   Using direct equality operators (== or ~=) may not be appropriate
%   for matrices of class DOUBLE or COMPLEX.  FLTEQ implements a weaker
%   sense of equality with exact equality definitions overridable by the
%   user.
%
%   Two objects, X and Y, are deemed equal if and only if
%    - ALL(SIZE(X) == SIZE(Y)), and
%    - (ALL(ABS(X - Y) < atol) or
%       ALL(ABS(X - Y) < rtol.*ABS(Y)))
%   with `atol' and `rtol' being absolute and relative tolerances
%   respectively.
%
%   For complex arrays, separate checks are made for the real and
%   imaginary parts.
%
% PARAMETERS:
%   x, y - Objects to check for equality.
%   atol - Absolute tolerance.  OPTIONAL.  DEFAULT VALUE = 1.0e-6.
%   rtol - Relative tolerance.  OPTIONAL.  DEFAULT VALUE = 1.0e-7.
%
% RETURNS:
%   B    - Boolean status indicating whether x is equal to y or not.
%          Possible values are  TRUE  and  FALSE.
%
% SEE ALSO:
%   RELOP, REAL, IMAG, TRUE, FALSE.

% DUPLICATION NOTE:
%
%   This function is an exact duplicate of FLOATEQUALS of the Expint
%   package.  The duplication is made in order to create a
%   self-contained PHIPADE function for use in other projects.  Any
%   change made to FLTEQ should be replicated in FLOATEQUALS if the
%   latter is available.

error(nargchk(2, 4, nargin));

if issparse(x) || issparse(y),
   % only work on non-zero elements...
   [ix, jx, x] = find(x);
   [iy, jy, y] = find(y);

   B = (numel(ix) == numel(iy)) && ...
       all(ix == iy) && all(jx == jy);
else
   B = true;                    % assume equality by default
end

sx = size(x);
sy = size(y);

if B && all(sx == sy),
   [atol, rtol] = deal(1.0e-6, 1.0e-7);

   if nargin > 2, atol = abs(varargin{1}); end
   if nargin > 3, rtol = abs(varargin{2}); end

   % Straighten out  x  and  y  for multi-D cases
   if all(sx > 1), x = x(:); y = y(:); end

   [xc, yc] = deal(real(x), real(y));
   xc = abs(xc - yc);
   a = all(xc < atol);
   r = all(xc < rtol.*abs(yc));

   if ~isreal(x) || ~isreal(y),
      [xc, yc] = deal(imag(x), imag(y));
      xc = abs(xc - yc);
      a = a & all(xc < atol);
      r = r & all(xc < rtol.*abs(yc));
   end

   B = a | r;
else
   B = false;
end
